unit USourceCode;

{$MODE Delphi}

interface

uses
  SysUtils, Variants, Classes,
  Dialogs, Contnrs;

type
  TMemoryType = (mtProgram, mtData, mtEEPROM);

  TModuleType = (mdProgram, mdUnit, mdLibrary, mdPackage);

  TNodeType = (ntNone, ntVariable, ntFunction, ntConstant, ntOperator);

  TTypeVisibility = (tvPublic, tvPublished, tvPrivate, tvProtected);

  TValue = Variant; //array of Byte;

  TCommonBlock = class;
  TTypeList = class;
  TConstantList = class;
  TVariableList = class;
  TFunctionList = class;
  TExpressionList = class;
  TExpression = class;
  TFunction = class;
  TVariable = class;
  TConstant = class;
  TModule = class;
  TProgram = class;

  TDevice = class
    Family: string;
    Memory: array[TMemoryType] of Integer;
  end;

  TContext = class

  end;

  TCommandList = class;

  TCommand = class
    Parent: TObject;
    CommonBlock: TCommonBlock;
  end;

  TAssignment = class(TCommand)
    Target: TVariable;
    Source: TExpression;
    constructor Create;
    destructor Destroy; override;
  end;

  { TFunctionCall }

  TFunctionCall = class(TCommand)
    FunctionRef: TFunction;
    ParameterExpression: TObjectList; // TObjectList<TExpression>
    constructor Create;
    destructor Destroy; override;
  end;

  TBeginEnd = class(TCommand)
    Commands: TCommandList;
    CommonBlock: TCommonBlock;
    procedure Clear;
    constructor Create;
    destructor Destroy; override;
  end;

  TWhileDo = class(TCommand)
    Condition: TExpression;
    Command: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  TWithDo = class(TCommand)
    Context: TContext;
    Command: TCommand;
  end;

  TRepeatUntil = class(TCommand)
    Block: TCommandList;
    Condition: TExpression;
  end;

  { TForToDo }

  TForToDo = class(TCommand)
    ControlVariable: TVariable;
    Start: TExpression;
    Stop: TExpression;
    Command: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  { TIfThenElse }

  TIfThenElse = class(TCommand)
    Condition: TExpression;
    Command: TCommand;
    ElseCommand: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  TCaseOfEndBranche = class
    Constant: TConstant;
    Command: TCommand;
  end;

  TCaseOfEnd = class(TCommand)
    Expression: TExpression;
    Branches: TObjectList; // TObjectList<TCaseOfEndBranche>
    ElseCommand: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  TTryFinally = class(TCommand)
    Block: TCommandList;
    FinallyBlock: TCommandList;
  end;

  TTryExcept = class(TCommand)
    Block: TCommandList;
    ExceptBlock: TCommandList;
  end;

  TCommandList = class(TObjectList)

  end;

  TCommonBlock = class
    Name: string;
    Parent: TCommonBlock;
    ParentModule: TModule;
    Constants: TConstantList;
    Types: TTypeList;
    Variables: TVariableList;
    Functions: TFunctionList;
    Code: TBeginEnd;
    constructor Create; virtual;
    destructor Destroy; override;
//    procedure CheckReferences;
  end;

  TType = class
    Internal: Boolean;
    Parent: TTypeList;
    Name: string;
    Size: Integer;
    UsedType: TType;
    Visibility: TTypeVisibility;
  end;

  TTypeInherited = class(TType)
  end;

  TTypeRecord = class(TType)
    CommonBlock: TCommonBlock;
    constructor Create;
    destructor Destroy; override;
  end;

  TTypeSubRange = class(TType)
    Bottom: Variant;
    Top: Variant;
  end;

  TTypeArray = class(TType)
    IndexType: TType;
    ItemType: TType;
  end;

  TTypePointer = class(TType)
  end;

  TEnumItem = class
    Name: string;
    Index: Integer;
  end;

  TTypeEnumeration = class(TType)
    Items: TObjectList; // TObjectList<TEnumItem>
    constructor Create;
    destructor Destroy; override;
  end;

  { TTypeClass }

  TTypeClass = class(TType)
    Items: TObjectList; // TObjectList<TType>
    constructor Create;
    destructor Destroy; override;
  end;

  TTypeList = class(TObjectList)
    Parent: TCommonBlock;
    function Search(Name: string; Exported: Boolean = False): TType;
    destructor Destroy; override;
  end;

  TConstant = class
    System: Boolean;
    Name: string;
    ValueType: TType;
    Value: TValue;
    Exported: Boolean;
  end;

  TConstantList = class(TObjectList)
    Parent: TCommonBlock;
    function Search(Name: string): TConstant;
    destructor Destroy; override;
  end;

  TVariable = class
    Name: string;
    ValueType: TType;
    DefaultValue: TValue;
    Exported: Boolean;
  end;

  TVariableList = class(TObjectList)
    Parent: TCommonBlock;
    function Search(Name: string; Exported: Boolean = False): TVariable;
    destructor Destroy; override;
  end;

  TParameter = class(TVariable)
  end;

  TParameterList = class(TObjectList)
    Parent: TFunction;
    function Search(Name: string): TParameter;
    destructor Destroy; override;
  end;

  { TExpression }

  TExpression = class
    CommonBlock: TCommonBlock;
    NodeType: TNodeType;
    Variable: TVariable;
    Method: TFunction;
    Value: TValue;
    OperatorName: string;
    SubItems: TExpressionList;
    Associated: Boolean;
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TExpression);
  end;

  TExpressionList = class(TObjectList)
    destructor Destroy; override;
  end;

  TFunctionType = (ftFunction, ftProcedure, ftConstructor, ftDestructor);

  TFunction = class(TCommonBlock)
  public
    Internal: Boolean;
    FunctionType: TFunctionType;
    Parameters: TParameterList;
    ResultType: TType;
    Exported: Boolean;
    constructor Create; override;
    destructor Destroy; override;
  end;

  TFunctionList = class(TObjectList)
    Parent: TCommonBlock;
    function Search(Name: string; Exported: Boolean = False): TFunction;
    destructor Destroy; override;
  end;

  TUsedModule = class
    Name: string;
    Location: string;
    Module: TModule;
    Exported: Boolean;
  end;

  TUsedModuleList = class(TObjectList)
    ParentModule: TModule;
  end;

  { TModule }

  TModule = class
  public
    ParentProgram: TProgram;
    Name: string;
    UsedModules: TUsedModuleList;
    Body: TCommonBlock;
    function SearchVariable(Name: string; Outside: Boolean): TVariable; virtual;
    function SearchConstant(Name: string; Outside: Boolean): TConstant; virtual;
    function SearchType(AName: string; Outside: Boolean): TType; virtual;
    function SearchFunction(Name: string; Outside: Boolean): TFunction; virtual;
    constructor Create;
    destructor Destroy; override;
  end;

  { TModuleProgram }

  TModuleProgram = class(TModule)
    constructor Create;
    destructor Destroy; override;
  end;

  { TModuleUnit }

  TModuleUnit = class(TModule)
    constructor Create;
    destructor Destroy; override;
  end;

  TModulePackage = class(TModule)

  end;

  TModuleLibrary = class(TModule)
  end;

  { TModuleList }

  TModuleList = class(TObjectList)
    function Search(Name: string): TModule;
  end;

  TProgram = class
    Device: TDevice;
    Modules: TModuleList;
    constructor Create;
    destructor Destroy; override;
  end;

const
  KeyWords: array[0..37] of string = ('program', 'unit', 'uses', 'begin', 'end',
    'type', 'const', 'var', 'array', 'record', 'absolute', 'virtual', 'class',
    'set', 'private', 'public', 'interface', 'implementation', 'finalization',
    'initialization', 'for', 'while', 'if', 'case', 'of', 'pointer',
    'object', 'packed', 'procedure', 'function', 'to', 'do', 'downto', 'repeat',
    'until', 'then', 'asm', 'else');
  Operators: array[0..22] of string = ('@', 'not', '*', 'and', '/', 'shl',
    'shr', 'as', 'div', 'mod', 'or', 'xor', '-', '+', '=', '>', '<', '<>', '<=',
     '>=', 'is', 'in', ':=');

implementation

{ TFunction }

constructor TFunction.Create;
begin
  inherited;
  Parameters := TParameterList.Create;
  Parameters.Parent := Self;
  //ResultType := TType.Create;
end;

destructor TFunction.Destroy;
begin
  Parameters.Free;
//  ResultType.Free;
  inherited;
end;

{ TProgram }

constructor TProgram.Create;
begin
  Device := TDevice.Create;
  Modules := TModuleList.Create;
end;

destructor TProgram.Destroy;
begin
  Modules.Free;
  Device.Free;
end;

{ TConstant }


{ TConstantList }

destructor TConstantList.Destroy;
begin
  inherited;
end;

function TConstantList.Search(Name: string): TConstant;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(TConstant(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TConstant(Items[I]) else begin
    if Assigned(Parent.Parent) then Result := Parent.Parent.Constants.Search(Name)
      else begin
        Result := nil;
      end;
  end;
end;

{ TModule }

constructor TModule.Create;
begin
  inherited;
  UsedModules := TUsedModuleList.Create;
  UsedModules.ParentModule := Self;
  Body := TCommonBlock.Create;
  Body.ParentModule := Self;
end;

destructor TModule.Destroy;
begin
  Body.Free;
  UsedModules.Free;
  inherited;
end;

(*
procedure TCommonBlock.CheckReferences;
var
  I: Integer;
begin
  for I := 0 to Operations.Count - 1 do
  with TOperation(Operations[I]) do begin
    if (Instruction = inJump) or (Instruction = inConditionalJump) then
      TOperation(Operations[GotoAddress]).Referenced := True;
  end;
end;
*)

constructor TCommonBlock.Create;
begin
  Constants := TConstantList.Create;
  Constants.Parent := Self;
  Types := TTypeList.Create;
  Types.Parent := Self;
  Variables := TVariableList.Create;
  Variables.Parent := Self;
  Functions := TFunctionList.Create;
  Functions.Parent := Self;
  Code := TBeginEnd.Create;
  Code.Parent := Self;
  Code.CommonBlock := Self;
end;

destructor TCommonBlock.Destroy;
begin
  Constants.Free;
  Types.Free;
  Variables.Free;
  Functions.Free;
  Code.Free;
  inherited;
end;

{ TTypeList }

destructor TTypeList.Destroy;
begin
  inherited;
end;

function TTypeList.Search(Name: string; Exported: Boolean = False): TType;
var
  I: Integer;
begin
  // Search in own list
  I := 0;
  while (I < Count) and (LowerCase(TType(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TType(Items[I]) else Result := nil;

  // Search in parent
  if not Assigned(Result) and Assigned(Parent.Parent) then
     Result := Parent.Parent.Types.Search(Name);

  // Search global variables
  if not Assigned(Result) and Assigned(Parent.ParentModule) and (not Exported) then begin
    Result := Parent.ParentModule.SearchType(Name, True);
  end;
end;

{ TVariableList }

destructor TVariableList.Destroy;
begin
  inherited;
end;

function TVariableList.Search(Name: string; Exported: Boolean = False): TVariable;
var
  I: Integer;
begin
  // Search in own list
  I := 0;
  while (I < Count) and (LowerCase(TVariable(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TVariable(Items[I]) else Result := nil;

  if not Exported then begin
    // Search parent block variables
    if not Assigned(Result) then begin
      if Assigned(Parent.Parent) then
        Result := Parent.Parent.Variables.Search(Name)
        else Result := nil;
    end;

    // Search parent function parameters
    if not Assigned(Result) then begin
      if Assigned(Parent) and (Parent is TFunction) then
        Result := TFunction(Parent).Parameters.Search(Name)
        else Result := nil;
    end;

    if not Assigned(Result) then begin
      if Assigned(Parent.Parent) and (Parent.Parent is TFunction) then
        Result := TFunction(Parent.Parent).Parameters.Search(Name)
        else Result := nil;
    end;

    // Search global variables
    //if not Assigned(Result) then begin
    //  if Assigned(Parent.ParentModule) then
    //    Result := Parent.ParentModule.SearchVariable(Name, True);
    //end;
  end;
end;

{ TFunctionList }

destructor TFunctionList.Destroy;
begin
  inherited;
end;

function TFunctionList.Search(Name: string; Exported: Boolean): TFunction;
var
  I: Integer;
begin
  // Search in own list
  I := 0;
  while (I < Count) and (LowerCase(TFunction(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TFunction(Items[I]) else Result := nil;

  // Search parent block list
  if not Assigned(Result) and Assigned(Parent.Parent) then
    Result := Parent.Parent.Functions.Search(Name, Exported);

  // Search global variables
  if not Assigned(Result) and Assigned(Parent.ParentModule) and (not Exported) then
    Result := Parent.ParentModule.SearchFunction(Name, True);
end;

{ TExpression }

constructor TExpression.Create;
begin
  SubItems := TExpressionList.Create;
  SubItems.Count := 2;
  SubItems.OwnsObjects := False;
end;

destructor TExpression.Destroy;
begin
  SubItems.Free;
  inherited;
end;

procedure TExpression.Assign(Source: TExpression);
begin
  CommonBlock := Source.CommonBlock;
  NodeType := Source.NodeType;
  Method := Source.Method;
  Value := Source.Value;
  Associated := Source.Associated;
  OperatorName := Source.OperatorName;
  Variable := Source.Variable;
  SubItems.Assign(Source.SubItems);
end;

{ TExpressionList }

destructor TExpressionList.Destroy;
begin
  inherited;
end;

{ TParameterList }

destructor TParameterList.Destroy;
begin
  inherited;
end;

function TParameterList.Search(Name: string): TParameter;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(TParameter(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TParameter(Items[I])
    else Result := nil;
end;

{ TBeginEnd }

procedure TBeginEnd.Clear;
begin

end;

constructor TBeginEnd.Create;
begin
  inherited;
  Commands := TCommandList.Create;
end;

destructor TBeginEnd.Destroy;
begin
  Commands.Free;
  inherited;
end;

{ TAssignment }

constructor TAssignment.Create;
begin
//  Source := TExpression.Create;
end;

destructor TAssignment.Destroy;
begin
  Source.Free;
  inherited;
end;

{ TWhileDo }

constructor TWhileDo.Create;
begin
  Condition := TExpression.Create;
  Command := TCommand.Create;
end;

destructor TWhileDo.Destroy;
begin
  Condition.Free;
  Command.Free;
  inherited;
end;

{ TCaseOfEnd }

constructor TCaseOfEnd.Create;
begin
  inherited;
  Branches := TObjectList.Create
end;

destructor TCaseOfEnd.Destroy;
begin
  Branches.Free;
  inherited;
end;

{ TIfThenElse }

constructor TIfThenElse.Create;
begin
  Condition := TExpression.Create;
end;

destructor TIfThenElse.Destroy;
begin
  Condition.Free;
  inherited Destroy;
end;

{ TFunctionCall }

constructor TFunctionCall.Create;
begin
  inherited;
  ParameterExpression := TObjectList.Create;
end;

destructor TFunctionCall.Destroy;
begin
  ParameterExpression.Free;
  inherited Destroy;
end;

{ TForToDo }

constructor TForToDo.Create;
begin
  inherited;
  Start := TExpression.Create;
  Stop := TExpression.Create;
end;

destructor TForToDo.Destroy;
begin
  Start.Free;
  Stop.Free;;
  inherited Destroy;
end;

{ TTypeRecord }

constructor TTypeRecord.Create;
begin
  inherited;
  CommonBlock := TCommonBlock.Create;
end;

destructor TTypeRecord.Destroy;
begin
  CommonBlock.Free;
  inherited Destroy;
end;

{ TModuleList }

function TModuleList.Search(Name: string): TModule;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(TModule(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TModule(Items[I]) else Result := nil;
end;

{ TModuleProgram }

function TModule.SearchVariable(Name: string; Outside: Boolean): TVariable;
var
  I: Integer;
begin
  Result := Body.Variables.Search(Name, Outside);

  if not Assigned(Result) and Outside then begin
    I := 0;
    while (I < UsedModules.Count) and (not Assigned(Result)) do begin
      Result := TUsedModule(UsedModules[I]).Module.SearchVariable(Name, False);
      Inc(I);
    end;
  end;
end;

function TModule.SearchConstant(Name: string; Outside: Boolean): TConstant;
begin

end;

function TModule.SearchType(AName: string; Outside: Boolean): TType;
var
  I: Integer;
begin
  Result := Body.Types.Search(AName, Outside);

  if not Assigned(Result) and Outside then begin
    I := 0;
    while (I < UsedModules.Count) and (not Assigned(Result)) do begin
      with TUsedModule(UsedModules[I]) do
        if Assigned(Module) then
        with Module do
          Result := SearchType(AName, False);
      Inc(I);
    end;
  end;
end;

function TModule.SearchFunction(Name: string; Outside: Boolean): TFunction;
var
  I: Integer;
begin
  Result := Body.Functions.Search(Name, Outside);

  if not Assigned(Result) and Outside then begin
    I := 0;
    while (I < UsedModules.Count) and (not Assigned(Result)) do begin
    if Assigned(TUsedModule(UsedModules[I]).Module) then
      Result := TUsedModule(UsedModules[I]).Module.SearchFunction(Name, False);
      Inc(I);
    end;
  end;
end;

constructor TModuleProgram.Create;
begin
  inherited;
end;

destructor TModuleProgram.Destroy;
begin
  inherited Destroy;
end;

{ TModuleUnit }


constructor TModuleUnit.Create;
begin
  inherited;
end;

destructor TModuleUnit.Destroy;
begin
  inherited Destroy;
end;

{ TTypeEnumeration }

constructor TTypeEnumeration.Create;
begin
  inherited;
  Items := TObjectList.Create;
end;

destructor TTypeEnumeration.Destroy;
begin
  Items.Free;
  inherited Destroy;
end;

{ TTypeClass }

constructor TTypeClass.Create;
begin
  Items := TObjectList.Create;
end;

destructor TTypeClass.Destroy;
begin
  Items.Free;
  inherited Destroy;
end;

end.

