unit UMainForm;

{$MODE Delphi}

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, UCompiler, UProducerAsm8051, Registry,
  UProducerDynamicC, ComCtrls, ExtCtrls, SynEdit, SynHighlighterPas, UProducerTreeView,
  UProducerPascal, Contnrs, UProject, FileUtil, Menus, ActnList, UCoolDocking,
  UCompiledForm, UCodeTreeForm, URegistry, ULastOpenedList, UApplicationInfo,
  UDebugLog;

const
  RegistryRootKey = HKEY_CURRENT_USER;
  RegistryKey = 'Software\Transpascal';

type
  { TMainForm }

  TMainForm = class(TForm)
    ABuild: TAction;
    AProjectNew: TAction;
    AProjectOpen: TAction;
    AProjectSave: TAction;
    AProjectSaveAs: TAction;
    AProjectClose: TAction;
    ActionList1: TActionList;
    Button1: TButton;
    CheckBoxLogParsing: TCheckBox;
    ComboBoxTargetSelection: TComboBox;
    CoolDockClient1: TCoolDockClient;
    CoolDockMaster1: TCoolDockMaster;
    MainMenu1: TMainMenu;
    MenuItem1: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem3: TMenuItem;
    MenuItem4: TMenuItem;
    MenuItem5: TMenuItem;
    MenuItem6: TMenuItem;
    MenuItem7: TMenuItem;
    MenuItem8: TMenuItem;
    DockPanel: TPanel;
    MenuItemOpenRecent: TMenuItem;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    Splitter3: TSplitter;
    ToolBar1: TToolBar;
    procedure AProjectCloseExecute(Sender: TObject);
    procedure AProjectOpenExecute(Sender: TObject);
    procedure AProjectSaveAsExecute(Sender: TObject);
    procedure ComboBox1Change(Sender: TObject);
    procedure ComboBoxTargetSelectionChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ABuildExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    LastOpenedFiles: TLastOpenedList;
    ReopenLastOpenedFile: Boolean;
    procedure CompilerDebugLog(Text: string);
    procedure OpenRecentClick(Sender: TObject);
    procedure DockInit;
    procedure LoadFromRegistry;
    procedure SaveToRegistry;
    procedure OpenProject(FileName: string);
  public
    Project: TProject;
    Compiler: TCompiler;
    function GetSource(Name: string; Source: TStringList): Boolean;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.lfm}

uses
  UMessagesForm, UCodeForm, UProjectManager;

{ TMainForm }

procedure TMainForm.ABuildExecute(Sender: TObject);
begin
  // Compile project file
  if CheckBoxLogParsing.Checked then
     Compiler.Parser.OnDebugLog := CompilerDebugLog
     else Compiler.Parser.OnDebugLog := nil;
  Compiler.ProducerType := TProducerType(ComboBoxTargetSelection.ItemIndex);
  Compiler.Init;
  Compiler.Parser.OnGetSource := GetSource;

  if Project.Items.Count > 0 then
  with TProjectFile(Project.Items[0]) do begin
    Compiler.TargetFolder := Project.RootDir;
    Compiler.Compile(Parent.GetDir + Name, Source);
  end;

  ProjectManager.TreeViewProjectChange(Self, ProjectManager.TreeViewProject.Selected);
  MessagesForm.Reload;
end;

function TMainForm.GetSource(Name: string; Source: TStringList): Boolean;
var
  FileName: string;
begin
  FileName := Project.GetDir(True) + Name + '.pas';
  if FileExists(FileName) then begin
    Source.LoadFromFile(FileName);
    Result := True;
  end else Result := False;
end;

procedure TMainForm.DockInit;
var
  Container1: TCoolDockConjoinForm;
  Container2: TCoolDockConjoinForm;
begin
  ProjectManager.ManualDock(DockPanel, nil, alLeft);
  ProjectManager.Show;

  Container1 := TCoolDockManager(DockPanel.DockManager).CreateContainer(alRight);
  Container1.Caption := 'Tools';
  Container1.Name := 'Tools';
  Container1.ManualDock(DockPanel, nil, alRight);

  CodeTreeForm.ManualDock(DockPanel, nil, alLeft);
  CodeTreeForm.Show;
  Container1.Parent.Parent.Width := MainForm.Width;

  Container2 := TCoolDockManager(DockPanel.DockManager).CreateContainer(alRight);
  Container2.Caption := 'Panel';
  Container2.Name := 'Tools2';
  Container2.ManualDock(Container1.Panel, nil, alRight);

  MessagesForm.ManualDock(Container1.Panel, nil, alBottom);
  MessagesForm.Show;

  CodeForm.ManualDock(Container2.Panel, nil, alTop);
  CodeForm.Show;

  CompiledForm.ManualDock(Container2.Panel, nil, alTop);
  CompiledForm.Show;

  TCoolDockManager(Container2.Panel.DockManager).DockStyle := dsTabs;

  ProjectManager.Parent.Parent.Width := 200;
  Container2.Parent.Parent.Height := MainForm.Height;
end;

procedure TMainForm.LoadFromRegistry;
begin
  with TRegistryEx.Create do
    try
      RootKey := HKEY_CURRENT_USER;
      OpenKey(RegistryKey, True);
      ReopenLastOpenedFile := ReadBoolWithDefault('ReopenLastOpenedFile', True);
      ComboBoxTargetSelection.ItemIndex :=
        ReadIntegerWithDefault('ProducerType', 0);
      CheckBoxLogParsing.Checked := ReadBoolWithDefault('LogParsing', False);
    finally
      Free;
    end;
  LastOpenedFiles.LoadFromRegistry(RegistryRootKey, RegistryKey);
end;

procedure TMainForm.SaveToRegistry;
begin
  with TRegistryEx.Create do
    try
      RootKey := RegistryRootKey;
      OpenKey(RegistryKey, True);
      WriteBool('ReopenLastOpenedFile', ReopenLastOpenedFile);
      WriteInteger('ProducerType', ComboBoxTargetSelection.ItemIndex);
      WriteBool('LogParsing', CheckBoxLogParsing.Checked);
    finally
      Free;
    end;
  LastOpenedFiles.SaveToRegistry(RegistryRootKey, RegistryKey);
  LastOpenedFiles.ReloadMenu;
end;

procedure TMainForm.OpenProject(FileName: string);
begin
  Project.LoadFromFile(FileName);
  LastOpenedFiles.Add(FileName);
  LastOpenedFiles.ReloadMenu;
  ProjectManager.UpdateProjectTree;
  Caption := ApplicationInfo.Name + ' - ' + Project.Name;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Project.Save;
  SaveToRegistry;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  ProducerType: TProducerType;
begin
  DebugLog.FileName := 'DebugLog.txt';
  DeleteFile(DebugLog.FileName);
  Compiler := TCompiler.Create;
  Project := TProject.Create;
  LastOpenedFiles := TLastOpenedList.Create;
  LastOpenedFiles.MenuItem := MenuItemOpenRecent;
  LastOpenedFiles.ClickAction := OpenRecentClick;

  ComboBoxTargetSelection.Clear;
  for ProducerType := Low(ProducerType) to High(ProducerType) do
    ComboBoxTargetSelection.AddItem(ProducerTypeName[ProducerType], nil);
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  LastOpenedFiles.Free;
  Project.Free;
  Compiler.Free;
end;

procedure TMainForm.CompilerDebugLog(Text: string);
begin
  DebugLog.Add('', Text);
end;

procedure TMainForm.OpenRecentClick(Sender: TObject);
begin
  if Sender is TMenuItem then
     OpenProject(StringReplace(TMenuItem(Sender).Caption, '&', '', [rfReplaceAll]));
end;

procedure TMainForm.FormShow(Sender: TObject);
var
  NewProjectFile: TProjectFile;
begin
  LoadFromRegistry;
  DockInit;

  if ReopenLastOpenedFile and (LastOpenedFiles.Count > 0) then begin
    OpenProject(LastOpenedFiles[0]);
  end;

  WindowState := wsMaximized;
end;

procedure TMainForm.ComboBox1Change(Sender: TObject);
begin
  ABuildExecute(Self);
end;

procedure TMainForm.AProjectOpenExecute(Sender: TObject);
begin
  if OpenDialog1.Execute then begin
    OpenProject(OpenDialog1.FileName);
  end;
end;

procedure TMainForm.AProjectCloseExecute(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.AProjectSaveAsExecute(Sender: TObject);
begin
  if SaveDialog1.Execute then begin
    Project.SaveToFile(SaveDialog1.FileName);
  end;
end;

procedure TMainForm.ComboBoxTargetSelectionChange(Sender: TObject);
begin
  ABuildExecute(Self);
end;

end.
